//
//------------------------------------------------------------------------------
//  Name    : I2CX Library
//  Author  : Jerry Messina
//  Date    : 17 FEB 2023
//  Version : 1.0
//
//  Software I2C Master library
//
//  I2CX is a bit-banged software I2C master that implements control
//  of the SCL/SDA pins in open-drain mode (so they both require a pullup).
//  It supports clock stretching and multi-master modes
//  Primarily for devices without an MSSP peripheral (ie xv18 w/I2C peripheral)
//
//  ver 1.0  17 FEB 2023
//      - Initial release
//------------------------------------------------------------------------------
//
module I2CX

include "system.bas"

// SCL pin option
#option I2CX_SCL = PORTC.3          // default SCL pin for most devices
#if not IsValidPortPin(I2CX_SCL)
  #error I2CX_SCL, "Invalid option. I2CX clock must be a valid port pin"
#endif

// SDA pin option
#option I2CX_SDA = PORTC.4          // default SDA pin for most devices
#if not IsValidPortPin(I2CX_SDA)
  #error I2CX_SDA, "Invalid option. I2CX data must be a valid port pin"
#endif

// I2C baudrate option (in KHz)
// actual rates are approx half the setting due to software overhead
#option I2CX_BAUD = 100
#if not (I2CX_BAUD in (100, 400, 1000))
  #error I2CX_BAUD, "Invalid option. I2CX_BAUD must be 100, 400, or 1000"
#endif

// I2C signal risetime, in usecs (open-drain low to high)
// adds a fixed delay to allow signals to slew high, varies with pullup value and bus capacitance
#option I2CX_OD_RISETIME = 1            // risetime
const OD_RISETIME = I2CX_OD_RISETIME

// slave device SDA output valid from SCL low (TAA)
#if (I2CX_BAUD = 100)
  #option I2CX_SDA_OUTPUT_VALID = 3     // typ 3.45us
#elseif (I2CX_BAUD = 400)
  #option I2CX_SDA_OUTPUT_VALID = 1     // typ 0.9us
#elseif (I2CX_BAUD = 1000) 
  #option I2CX_SDA_OUTPUT_VALID = 0     // typ 0.18us
#endif
const SDA_OUTPUT_VALID = I2CX_SDA_OUTPUT_VALID

// slave device bus free time between operations (TBUF)
#if (I2CX_BAUD = 100)
  #option I2CX_BUS_FREE_TIME = 4        // typ 4.7us
#elseif (I2CX_BAUD = 400)
  #option I2CX_BUS_FREE_TIME = 1        // typ 1.3us
#elseif (I2CX_BAUD = 1000) 
  #option I2CX_BUS_FREE_TIME = 0        // typ 0
#endif
const BUS_FREE_TIME = I2CX_BUS_FREE_TIME

// I2C clock arbitration (clock stretching)
//  true = wait for clock to be seen high (default)
//  false = wait fixed OD_RISETIME
#option I2CX_CLOCK_ARB = true
#if not (I2CX_CLOCK_ARB in (true, false))
  #error I2CX_CLOCK_ARB, "Invalid option. I2CX_CLOCK_ARB must be true or false"
#endif  

// optimize code for speed vs size
//  true = faster, more code
//  false = slower, less code (default)
#option I2CX_OPT_SPEED = false
#if not (I2CX_OPT_SPEED in (true, false))
  #error I2CX_OPT_SPEED, "Invalid option. I2CX_OPT_SPEED must be true or false"
#endif  

// include user init callback event
//   false = callback event not used (saves ram/code, default)
//   true = use init callback event, set/called during Initialize()
// can be used to setup pullups, digital mode, etc 
#option I2CX_INIT_EVENT = false
#if (I2CX_INIT_EVENT)
type event_t = event()
public dim
    OnI2CXInit as event_t = null        // user event callback for I2C init
#endif

// allow IsPresent to detect addresses with R/W=1 (addr bit 0)
#option I2CX_PROBE_RD_ADDR = false
#if not (I2CX_PROBE_RD_ADDR in (true, false))
  #error I2CX_PROBE_RD_ADDR, "Invalid option. I2CX_PROBE_RD_ADDR must be true or false"
#endif  

// define SCL and SDA pins from options
dim
    SCL as I2CX_SCL.I2CX_SCL@,
    SDA as I2CX_SDA.I2CX_SDA@
   
// acknowledge data bit (during 9th clock pulse)
// the ACK bit is set by the receiver... slave for WriteByte, master for ReadByte
public const
    I2C_ACKNOWLEDGE = 0,                     // low = acknowledge
    I2C_NOT_ACKNOWLEDGE = 1,                 // high = NOT acknowledge
    I2C_ACK = I2C_ACKNOWLEDGE,
    I2C_NACK = I2C_NOT_ACKNOWLEDGE

public structure I2C_stat_t
    b as byte
    ackstat as b.bits(0)        // ack bit
    BCL     as b.bits(1)        // bus collision
    WCOL    as b.bits(2)        // write collision
end structure

public dim
    I2C_stat as I2C_stat_t,
    NotAcknowledged as I2C_stat.b.booleans(0)   // for compat with old module

//
//--------------------------------------------------------------------------
// IO controls and delay routines
//--------------------------------------------------------------------------
//
// Tbrg baud rate generator delay (SCL low/high time)
#if (I2CX_OPT_SPEED)
inline 
#endif
sub TBRG_DELAY()
  #if (I2CX_BAUD = 100)
    delayus(5)
  #elseif (I2CX_BAUD = 400)
    delayus(1)      // shorten wait for code overhead (400K=2.5us)
  #else
    nop()
  #endif
end sub

// delay for low->high transitions in open-drain mode
#if (I2CX_OPT_SPEED)
inline 
#endif
sub OD_RISETIME_DELAY()
  #if (I2CX_OD_RISETIME > 0)
    delayus(OD_RISETIME)
  #else
    nop()       // allow at least one Tcy
  #endif
end sub

// SDA to SCL setup time
// allow at least 250ns data setup time before clk rising edge
// (does not account for risetime)
inline sub DATA_SETUP_DELAY()
    nop()
  #if (_clock >= 16)
    nop()
  #endif
  #if (_clock >= 32)
    nop()
    nop()
  #endif
end sub

// slave device SDA output valid from SCL low (TAA)
#if (I2CX_OPT_SPEED)
inline 
#endif
sub OUTPUT_VALID_FROM_CLOCK()
  #if (I2CX_SDA_OUTPUT_VALID > 0)
    delayus(SDA_OUTPUT_VALID)
  #else
    DATA_SETUP_DELAY()          // allow time for SDA to slew...
  #endif
end sub
    
// wait for external device to release the clock
#if (I2CX_OPT_SPEED)
inline 
#endif
sub WAIT_UNTIL_SCL_HIGH()
    while (SCL = 0)
    end while
end sub

// SCL high
#if (I2CX_OPT_SPEED) 
inline 
#endif
sub SCL_HIGH()
    input(SCL)              // input for open-drain mode... pulled high
    OD_RISETIME_DELAY()     // wait for risetime
  #if (I2CX_CLOCK_ARB)
    WAIT_UNTIL_SCL_HIGH()   // clock arbitration
  #endif
    TBRG_DELAY()
end sub

// SCL low
#if (I2CX_OPT_SPEED)
inline 
#endif
sub SCL_LOW()
    low(SCL)
    TBRG_DELAY()
end sub

// SDA high
inline sub SDA_HIGH()
    input(SDA)              // input for open-drain mode... pulled high
    OD_RISETIME_DELAY()     // wait for risetime
end sub

// SDA low
inline sub SDA_LOW()
    low(SDA)
end sub

//
//--------------------------------------------------------------------------
// ClearStatus
//  clear status reg
//--------------------------------------------------------------------------
//
public inline sub ClearStatus()
    I2C_stat.b = 0
end sub

//
//--------------------------------------------------------------------------
// Initialize
//  init I2C bus IO
//--------------------------------------------------------------------------
//
public sub Initialize()
  #if (I2CX_INIT_EVENT)
    // call user init event (set pins to digital mode, enable pullups, etc)
    OnI2CXInit()
  #endif
  
    // release SCL and SDA
    // they should both go high due to the pullups on the bus
    input(SCL)
    input(SDA)
    TBRG_DELAY()    // allow time for IO to go high
    
    ClearStatus()
end sub

#if (I2CX_INIT_EVENT)
public sub Initialize(pOnI2CXInit as event_t)
    OnI2CXInit = pOnI2CXInit
    Initialize()
end sub
#endif

//
//--------------------------------------------------------------------------
// Open/Close module
//  enables/disables I2C
//--------------------------------------------------------------------------
//         
// alias Open() to Initialize()
public dim Open as Initialize

// disable 
public sub Close()
    input(SCL)      // release control of the bus IO pins
    input(SDA)
end sub

//
//--------------------------------------------------------------------------
// CheckBus
//  check state of SDA and SCL
// returns : 
//  true = SDA and SCL are both high (bus ok)
//  false = SCL or SDA is low
//--------------------------------------------------------------------------
//         
public function CheckBus() as boolean
    // assume SDA and SCL are both high (bus ok)
    result = true

    // if either SCL or SDA are low return false
    if (SCL = 0) or (SDA = 0) then
        result = false
    endif
end function

//
//--------------------------------------------------------------------------
// Start
//  Send an I2C bus start condition (SDA HIGH->LOW while SCL HIGH)
//--------------------------------------------------------------------------
//
public sub Start()
    // bus must be idle before asserting start
    if (CheckBus() = false) then
        I2C_stat.BCL = 1
        exit
    endif
    
    SDA_HIGH()
    SCL_HIGH()      // SCL=1, SDA=1 (idle)
    SDA_LOW()       // SDA=1->0... start
    TBRG_DELAY()
    SCL_LOW()
end sub

//
//--------------------------------------------------------------------------
// Stop
//  Send an I2C bus stop condition (SDA HIGH->LOW while SCL HIGH)
//--------------------------------------------------------------------------
//
public sub Stop()
    SDA_LOW()       // assert SDA low
    TBRG_DELAY()    // wait one Tbrg
    SCL_HIGH()
    SDA_HIGH()      // SDA=0->1... stop
    TBRG_DELAY()
    // wait bus free time between operations
  #if (I2CX_BUS_FREE_TIME > 0)
    delayus(BUS_FREE_TIME)
  #endif    
end sub

//
//--------------------------------------------------------------------------
// Restart (Repeated Start)
//  Send an I2C bus repeated start (SDA HIGH->LOW while SCL HIGH)
//  A Restart is valid any time that a Stop would be valid
//  A master can issue a Restart if it wishes to hold the bus after terminating the current transfer
//  Restart has the same effect on the slave as a Start
//--------------------------------------------------------------------------
//
public sub Restart()
    SCL_LOW()
    SDA_HIGH()      // release SDA
    TBRG_DELAY()    // wait one Tbrg
    SCL_HIGH()      // SDA=1, SCL=1 (idle)
    SDA_LOW()       // SDA=1->0... repeated start
    TBRG_DELAY()
    SCL_LOW()
end sub

//
//--------------------------------------------------------------------------
// Acknowledge (master receiver)
//  Initiate I2C acknowledge
// inputs:
//  pAck = 1 (NOT acknowledge), 
//       = 0 (acknowledge)
//--------------------------------------------------------------------------
//         
public sub Acknowledge(pAck as bit = I2C_ACKNOWLEDGE)
    // 9th bit... send ack bit back to slave transmitter
    if (pAck = 1) then
        SDA_HIGH()
    else
        SDA_LOW()
    endif
    DATA_SETUP_DELAY()
    SCL_HIGH()
    SCL_LOW()

    SDA_LOW()       // exit with SDA=0, SCL=0
    SDA = pAck
end sub

//
//--------------------------------------------------------------------------
// ReadByte (master receiver)
//  Read a byte from the I2C bus WITHOUT sending ack bit
//  note: user must follow this function with a call to Acknowledge()
//      to complete the read cycle
// returns:
//  byte read from slave
//--------------------------------------------------------------------------
//
public function ReadByte() as byte
    dim count as byte

    input(SDA)      // release SDA
    result = 0
    count = 8
    repeat
        SCL_HIGH()
        result = result << 1
        result.bits(0) = SDA
        SCL_LOW()
        OUTPUT_VALID_FROM_CLOCK()
        dec(count)
    until (count = 0)
    ClrWDT()

    // complete the cycle using Acknowledge()
end function

//
//--------------------------------------------------------------------------
// ReadByte (master receiver)
//  Read a byte from the I2C bus, with Acknowledge
// returns:
//  byte read from slave
//--------------------------------------------------------------------------
//
public function ReadByte(pAck as bit) as byte
    result = ReadByte()
    Acknowledge(pAck)
end function

//
//--------------------------------------------------------------------------
// WriteByte (master transmitter)
//  Write a byte to the I2C bus
// returns : 
//  0 = ACK
//  1 = NACK
//--------------------------------------------------------------------------
//
public function WriteByte(b as byte) as bit
    dim count as byte

    count = 8
    repeat
        if (b.bits(7) = 1) then     // output data MSB to LSB
            SDA_HIGH()
        else
            SDA_LOW()
        endif            
        DATA_SETUP_DELAY()
        b = b << 1                  // get next bit (adds addtl data setup)
        SCL_HIGH()
        SCL_LOW()
        count = count - 1
    until (count = 0)

    // get ack bit...
    input(SDA)                  // release SDA so slave can ACK
    OUTPUT_VALID_FROM_CLOCK()
    SCL_HIGH()
    WAIT_UNTIL_SCL_HIGH()       // allow for slave clock stretching
    if (SDA = 0) then           // read slave ACK...
        result = I2C_ACK        // we got an ACK
    else
        result = I2C_NACK       // slave NACK
    endif
    I2C_stat.ackstat = result
    SCL_LOW()                   // exit with SCL=0, SDA=0
    SDA_LOW()

    ClrWDT()
end function

//
//--------------------------------------------------------------------------
// AckPoll (acknowledge polling)
//  Poll an addr until it returns an ACK or timeout
//  Can be used as an EEPROM write busy check (with a 5-10ms typ timeout), 
//  or a device presence check (leave timeout=0)
// input:
//  addr = I2C address ("control byte")
//  timeout = poll timeout, in msecs (0=single check, do not wait)
// returns:
//  true = addr ACKed within the timeout period
//  false= no ACK/timeout
//--------------------------------------------------------------------------
//
public function AckPoll(addr as byte, timeout as byte=0) as boolean
    const POLL_INTERVAL = 100   // interval for polling, in usecs (1000 max)
    dim tout as word            // timeout increments
    
    // convert msec timeout parameter to POLL_INTERVAL count
    // (typ faster than just using 1ms delay intervals)
    tout = timeout * (1000/POLL_INTERVAL)

    addr.bits(0) = 0        // ack polling uses RD_WRN = 0
    // send addr byte until device returns ACK (or timeout)
    repeat
        Start()
        WriteByte(addr)         // send addr and check for ack  
        Stop()
        if (I2C_stat.ackstat = I2C_ACK) then
            result = true
            exit
        endif            
        if (tout > 0) then
            delayus(POLL_INTERVAL)
            tout = tout - 1
        endif
    until (tout = 0)

    // device did not ack its addr within the timeout
    result = false
end function

// alias
public dim WaitForWrite as AckPoll

//
//--------------------------------------------------------------------------
// IsPresent
//  Detect if a slave addr is present
// returns:
//  true = slave acks addr (present)
//  false= slave nack (not present)
//--------------------------------------------------------------------------
//
public function IsPresent(addr as byte) as boolean
  #if (I2CX_PROBE_RD_ADDR = false)
    // bit 0 of the addr byte is the R/W bit and not technically
    // part of the address. this probes only "write" addresses
    addr.bits(0) = 0
  #endif

    // send the slave address and see if it acks
    result = false      // assume slave will nack... not detected
    Start()
    if (WriteByte(addr) = I2C_ACK) then
        result = true   // slave ack'd the addr
    endif
    Stop()
end function

//
//--------------------------------------------------------------------------
// ClearBus
//  clears a stuck bus
// returns : 
//  true = bus clear
//  false = bus still stuck
//--------------------------------------------------------------------------
//
// If the data line (SDA) is stuck LOW, the master should send nine clock 
// pulses, and the device that held the bus LOW should release it sometime 
// within those nine clocks. If not, reset or cycle power to clear the bus
// In the event the clock (SCL) is stuck LOW, the preferential procedure 
// is to reset the bus using a HW reset signal if your device has one, 
// otherwise you must cycle power to the device to clear the fault.
//
public function ClearBus() as boolean
    dim count as byte

    // disable bus so we can take control of the pins
    Close()

    delayus(5)
    // check SCL... if it's low then you're stuck so exit
    if (SCL = 0) then
        I2C_stat.BCL = 1
        result = false
        exit
    endif

    // send 9 clock pulses
    count = 9
    repeat
        SCL_LOW()
        delayus(5)
        SCL_HIGH()
        delayus(5)
        count = count - 1
    until (count = 0)

    // check to see if SCL and SDA are now high
    // if not you're stuck so exit
    if (CheckBus() = false) then
        I2C_stat.BCL = 1
        result = false
        exit
    endif

    // we managed to clear the fault...
    // reenable the bus and signal the result
    Open()
    result = true
end function

//
//--------------------------------------------------------------------------
// module initialization
//--------------------------------------------------------------------------
//

end module
